from __future__ import annotations

from datetime import datetime
from time import sleep
from typing import TYPE_CHECKING

import pytest
from sqlalchemy.exc import IntegrityError, NoResultFound

from bt_ddos_shield.state_manager import (
    AbstractMinerShieldStateManager,
    MinerShieldState,
    SQLAlchemyMinerShieldStateManager,
)
from bt_ddos_shield.utils import ShieldAddress

if TYPE_CHECKING:
    from bt_ddos_shield.utils import Hotkey, PublicKey
    from tests.conftest import ShieldTestSettings


class MemoryMinerShieldStateManager(AbstractMinerShieldStateManager):
    def __init__(self):
        super().__init__()
        self.current_miner_shield_state = MinerShieldState(
            known_validators={},
            banned_validators={},
            validators_addresses={},
            address_manager_state={},
            address_manager_created_objects={},
        )

    def add_validator(self, validator_hotkey: Hotkey, validator_public_key: PublicKey, redirect_address: ShieldAddress):
        self._state_add_validator(validator_hotkey, validator_public_key, redirect_address)

    def update_validator_public_key(self, validator_hotkey: Hotkey, validator_public_key: PublicKey):
        self._state_update_validator_public_key(validator_hotkey, validator_public_key)

    def add_banned_validator(self, validator_hotkey: Hotkey):
        assert self.current_miner_shield_state is not None
        if validator_hotkey in self.current_miner_shield_state.banned_validators:
            return
        self._state_add_banned_validator(validator_hotkey, datetime.now())

    def remove_validator(self, validator_hotkey: Hotkey):
        self._state_remove_validator(validator_hotkey)

    def update_address_manager_state(self, key: str, value: str | None):
        self._state_update_address_manager_state(key, value)

    def add_address_manager_created_object(self, obj_type: str, obj_id: str):
        self._state_add_address_manager_created_object(obj_type, obj_id)

    def del_address_manager_created_object(self, obj_type: str, obj_id: str):
        self._state_del_address_manager_created_object(obj_type, obj_id)

    def _load_state_from_storage(self) -> MinerShieldState:
        assert self.current_miner_shield_state is not None
        return self.current_miner_shield_state


class TestMinerShieldStateManager:
    """
    Test suite for the state manager.
    """

    @classmethod
    def create_db_state_manager(cls, test_settings: ShieldTestSettings) -> SQLAlchemyMinerShieldStateManager:
        state_manager = SQLAlchemyMinerShieldStateManager(test_settings.sql_alchemy_db_url)
        state_manager.clear_tables()
        state_manager.get_state()
        return state_manager

    def test_active_validators(self, shield_settings: ShieldTestSettings):
        validator1_hotkey = 'validator1'
        validator1_publickey = 'publickey1'
        validator1_address = ShieldAddress(address_id='validator1_id', address='1.2.3.4', port=80)

        validator2_hotkey = 'validator2'
        validator2_publickey = 'publickey2'
        validator2_new_publickey = 'new_publickey2'
        validator2_address = ShieldAddress(address_id='validator2_id', address='2.3.4.5', port=81)

        state_manager = self.create_db_state_manager(shield_settings)

        state_manager.add_validator(validator1_hotkey, validator1_publickey, validator1_address)
        # can't add again same address
        with pytest.raises(IntegrityError):
            state_manager.add_validator(validator2_hotkey, validator2_publickey, validator1_address)
        state_manager.add_validator(validator2_hotkey, validator2_publickey, validator2_address)
        # can't add again same validator
        with pytest.raises(IntegrityError):
            state_manager.add_validator(validator1_hotkey, validator1_publickey, validator1_address)

        state_manager.update_validator_public_key(validator2_hotkey, validator2_new_publickey)

        state_manager.remove_validator(validator1_hotkey)
        with pytest.raises(NoResultFound):
            state_manager.remove_validator(validator1_hotkey)

        with pytest.raises(NoResultFound):
            state_manager.update_validator_public_key(validator1_hotkey, validator2_new_publickey)

        state: MinerShieldState = state_manager.get_state()
        assert state.known_validators == {validator2_hotkey: validator2_new_publickey}
        assert state.validators_addresses == {validator2_hotkey: validator2_address}

        reloaded_state: MinerShieldState = state_manager.get_state(reload=True)
        assert state == reloaded_state

    def test_banned_validators(self, shield_settings: ShieldTestSettings):
        banned_validator_hotkey = 'banned_validator'

        state_manager = self.create_db_state_manager(shield_settings)

        state_manager.add_banned_validator(banned_validator_hotkey)
        ban_time: datetime = state_manager.get_state().banned_validators[banned_validator_hotkey]
        sleep(2)
        state_manager.add_banned_validator(banned_validator_hotkey)
        assert ban_time == state_manager.get_state().banned_validators[banned_validator_hotkey], (
            'first ban time should not change'
        )

        state: MinerShieldState = state_manager.get_state()
        assert state.banned_validators == {banned_validator_hotkey: ban_time}

        reloaded_state: MinerShieldState = state_manager.get_state(reload=True)
        assert state == reloaded_state

    def test_address_manager_state(self, shield_settings: ShieldTestSettings):
        key1 = 'key1'
        value1 = 'value1'
        key2 = 'key2'
        value2 = 'value2'
        key3 = 'key3'

        state_manager = self.create_db_state_manager(shield_settings)

        # Add key-value pairs to the address manager state
        state_manager.update_address_manager_state(key1, value1)
        state_manager.update_address_manager_state(key2, value2)

        state: MinerShieldState = state_manager.get_state()
        assert state.address_manager_state == {key1: value1, key2: value2}

        # Update an existing key
        new_value1 = 'new_value1'
        state_manager.update_address_manager_state(key1, new_value1)
        state = state_manager.get_state()
        assert state.address_manager_state == {key1: new_value1, key2: value2}

        # Remove a key
        state_manager.update_address_manager_state(key2, None)
        state = state_manager.get_state()
        assert state.address_manager_state == {key1: new_value1}

        # Ensure a non-existent key is not present
        state_manager.update_address_manager_state(key3, None)
        state = state_manager.get_state()
        assert key3 not in state.address_manager_state

        reloaded_state: MinerShieldState = state_manager.get_state(reload=True)
        assert state == reloaded_state

    def test_address_manager_created_objects(self, shield_settings: ShieldTestSettings):
        object_type1 = 'type1'
        object_id1 = 'id1'
        object_type2 = 'type2'
        object_id2 = 'id2'
        object_id3 = 'id3'

        state_manager = self.create_db_state_manager(shield_settings)

        # Add objects to the address manager created objects
        state_manager.add_address_manager_created_object(object_type1, object_id1)
        state_manager.add_address_manager_created_object(object_type2, object_id2)

        state: MinerShieldState = state_manager.get_state()
        assert state.address_manager_created_objects == {object_type1: {object_id1}, object_type2: {object_id2}}

        # Add another object to an existing type
        state_manager.add_address_manager_created_object(object_type2, object_id3)
        state = state_manager.get_state()
        assert state.address_manager_created_objects == {
            object_type1: {object_id1},
            object_type2: {object_id2, object_id3},
        }

        reloaded_state: MinerShieldState = state_manager.get_state(reload=True)
        assert state == reloaded_state

        # Remove an object
        state_manager.del_address_manager_created_object(object_type2, object_id2)
        state = state_manager.get_state()
        assert state.address_manager_created_objects == {object_type1: {object_id1}, object_type2: {object_id3}}

        # Ensure a non-existent object is not present
        state_manager.del_address_manager_created_object(object_type2, 'non_existent_id')
        state = state_manager.get_state()
        assert state.address_manager_created_objects == {object_type1: {object_id1}, object_type2: {object_id3}}

        # Remove last object of given type
        state_manager.del_address_manager_created_object(object_type2, object_id3)
        state = state_manager.get_state()
        assert state.address_manager_created_objects == {object_type1: {object_id1}}

        reloaded_state = state_manager.get_state(reload=True)
        assert state == reloaded_state



from __future__ import annotations

import asyncio
from time import sleep
from typing import TYPE_CHECKING

from bt_ddos_shield.encryption_manager import ECIESEncryptionManager
from bt_ddos_shield.event_processor import PrintingMinerShieldEventProcessor
from bt_ddos_shield.miner_shield import MinerShield, MinerShieldFactory, MinerShieldOptions
from bt_ddos_shield.shield_metagraph import ShieldMetagraph
from bt_ddos_shield.state_manager import MinerShieldState, SQLAlchemyMinerShieldStateManager
from bt_ddos_shield.validators_manager import (
    BittensorValidatorsManager,
    MemoryValidatorsManager,
)
from tests.test_address_manager import MemoryAddressManager
from tests.test_blockchain_manager import MemoryBlockchainManager
from tests.test_manifest_manager import MemoryManifestManager
from tests.test_state_manager import MemoryMinerShieldStateManager

if TYPE_CHECKING:
    import bittensor_wallet

    from bt_ddos_shield.address_manager import AbstractAddressManager
    from bt_ddos_shield.blockchain_manager import AbstractBlockchainManager
    from bt_ddos_shield.manifest_manager import AbstractManifestManager, Manifest
    from bt_ddos_shield.utils import Hotkey, PublicKey
    from tests.conftest import ShieldTestSettings


def generate_test_public_key() -> PublicKey:
    return ECIESEncryptionManager.serialize_certificate(ECIESEncryptionManager.generate_certificate()).public_key


class TestMinerShield:
    """
    Test suite for the MinerShield class.
    """

    MINER_HOTKEY: Hotkey = 'miner'
    VALIDATOR_1_HOTKEY: Hotkey = 'validator1'
    VALIDATOR_1_PUBLICKEY: PublicKey = generate_test_public_key()
    VALIDATOR_2_HOTKEY: Hotkey = 'validator2'
    VALIDATOR_2_PUBLICKEY: PublicKey = generate_test_public_key()
    VALIDATOR_3_HOTKEY: Hotkey = 'validator3'
    VALIDATOR_3_PUBLICKEY: PublicKey = generate_test_public_key()
    VALIDATOR_3_OTHER_PUBLICKEY: PublicKey = generate_test_public_key()
    VALIDATOR_4_HOTKEY: Hotkey = 'validator4'
    VALIDATOR_4_PUBLICKEY: PublicKey = generate_test_public_key()
    OTHER_VALIDATOR_HOTKEY: Hotkey = 'other_validator'
    DEFAULT_VALIDATORS = {
        VALIDATOR_1_HOTKEY: VALIDATOR_1_PUBLICKEY,
        VALIDATOR_2_HOTKEY: VALIDATOR_2_PUBLICKEY,
        VALIDATOR_3_HOTKEY: VALIDATOR_3_PUBLICKEY,
    }

    @classmethod
    def create_memory_validators_manager(
        cls, validators: dict[Hotkey, PublicKey] | None = None
    ) -> MemoryValidatorsManager:
        if validators is None:
            validators = cls.DEFAULT_VALIDATORS
        return MemoryValidatorsManager(dict(validators))

    def create_default_shield(self, validate_interval_sec: int = 2):
        self.validators_manager: MemoryValidatorsManager = self.create_memory_validators_manager()
        self.address_manager: MemoryAddressManager = MemoryAddressManager()
        self.manifest_manager: MemoryManifestManager = MemoryManifestManager()
        self.blockchain_manager: MemoryBlockchainManager = MemoryBlockchainManager(self.MINER_HOTKEY)
        self.state_manager: MemoryMinerShieldStateManager = MemoryMinerShieldStateManager()
        self.shield = MinerShield(
            self.validators_manager,
            self.address_manager,
            self.manifest_manager,
            self.blockchain_manager,
            self.state_manager,
            PrintingMinerShieldEventProcessor(),
            MinerShieldOptions(retry_delay_sec=1, validate_interval_sec=validate_interval_sec),
        )
        self.shield.enable()
        assert self.shield.run

    def test_start_stop_with_exception(self):
        """
        Test if shield is properly starting and stopping when exception is thrown during initialization.
        """
        state_manager = None  # set state_manager to None to force exception during initialization

        # noinspection PyTypeChecker
        shield = MinerShield(
            self.create_memory_validators_manager(),
            MemoryAddressManager(),
            MemoryManifestManager(),
            MemoryBlockchainManager(self.MINER_HOTKEY),
            state_manager,
            PrintingMinerShieldEventProcessor(),
            MinerShieldOptions(retry_delay_sec=1),
        )
        shield.enable()
        assert shield.run
        sleep(1)
        shield.disable()
        assert not shield.run

    def test_full_flow(self) -> None:
        """
        Test if shield is properly starting from scratch and fully enabling protection using mock memory managers.
        """
        self.create_default_shield()
        sleep(2 + 2 * self.shield.options.validate_interval_sec)  # Give some time to make sure validate is called

        try:
            state: MinerShieldState = self.state_manager.get_state()
            assert self.validators_manager.get_validators() == self.DEFAULT_VALIDATORS
            assert state.known_validators == self.validators_manager.get_validators()
            assert state.banned_validators == {}
            assert state.validators_addresses.keys() == self.validators_manager.get_validators().keys()
            assert list(state.validators_addresses.values()) == list(self.address_manager.known_addresses.values())
            assert self.address_manager.id_counter == len(state.validators_addresses)
            assert self.manifest_manager.stored_file is not None
            assert self.manifest_manager.put_counter == 1
            manifest_url: str = self.manifest_manager.get_manifest_url()
            manifest: Manifest = asyncio.run(self.manifest_manager.get_manifest(manifest_url))
            assert manifest.encrypted_url_mapping.keys() == state.validators_addresses.keys()
            assert asyncio.run(self.blockchain_manager.get_own_manifest_url()) == manifest_url
            assert self.blockchain_manager.put_counter == 1
        finally:
            self.shield.disable()
            assert not self.shield.run

    def test_integration(self, shield_settings: ShieldTestSettings):
        """
        Test if shield is properly starting from scratch and fully enabling protection using real managers.

        IMPORTANT: Test can run for many minutes due to AWS delays.
        """

        shield: MinerShield = MinerShieldFactory.create_miner_shield(shield_settings)

        assert isinstance(shield.state_manager, SQLAlchemyMinerShieldStateManager)
        state_manager: SQLAlchemyMinerShieldStateManager = shield.state_manager  # type: ignore
        state_manager.clear_tables()

        assert isinstance(shield.validators_manager, BittensorValidatorsManager)
        validators_manager: BittensorValidatorsManager = shield.validators_manager  # type: ignore

        manifest_manager: AbstractManifestManager = shield.manifest_manager
        address_manager: AbstractAddressManager = shield.address_manager

        validator_wallet: bittensor_wallet.Wallet = shield_settings.validator_wallet.instance
        validator_hotkey: Hotkey = validator_wallet.hotkey.ss58_address

        miner_wallet: bittensor_wallet.Wallet = shield_settings.wallet.instance
        miner_hotkey: Hotkey = miner_wallet.hotkey.ss58_address

        metagraph: ShieldMetagraph = ShieldMetagraph(
            wallet=validator_wallet,
            certificate_path=shield_settings.validator_cert_path,
            subtensor=shield_settings.subtensor.create_client(),
            netuid=shield_settings.netuid,
        )
        blockchain_manager: AbstractBlockchainManager = metagraph.blockchain_manager

        # Shorten validate_interval_sec and retry_delay_sec to speed up tests
        shield.options.validate_interval_sec = 10
        shield.options.retry_delay_sec = 1

        shield.enable()
        assert shield.run

        # Wait for initialization and ...
        shield.task_queue.join()
        # ... make sure validate is called
        sleep(2 * shield.options.validate_interval_sec)

        try:
            state: MinerShieldState = state_manager.get_state()
            assert validators_manager.get_validators() == {validator_hotkey: metagraph.certificate.public_key}
            assert state.known_validators == validators_manager.get_validators()
            assert state.banned_validators == {}
            assert state.validators_addresses.keys() == validators_manager.get_validators().keys()
            manifest_url: str = manifest_manager.get_manifest_url()
            manifest: Manifest = asyncio.run(manifest_manager.get_manifest(manifest_url))
            assert manifest.encrypted_url_mapping.keys() == state.validators_addresses.keys()
            urls: dict[Hotkey, str | None] = asyncio.run(blockchain_manager.get_manifest_urls([miner_hotkey]))
            assert urls[miner_hotkey] == manifest_url

            reloaded_state: MinerShieldState = state_manager.get_state(reload=True)
            assert reloaded_state == state

            # Set unknown validator manually to clean up everything (except S3 file) ...
            validators_manager.validators = frozenset(['unknown_hotkey'])
            # ... and wait for validate loop to apply changes
            sleep(2 * shield.options.validate_interval_sec)

            state = state_manager.get_state()
            assert state.known_validators == {}
            assert state.validators_addresses == {}
            manifest = asyncio.run(manifest_manager.get_manifest(manifest_url))
            assert manifest.encrypted_url_mapping == {}
            urls = asyncio.run(blockchain_manager.get_manifest_urls([miner_hotkey]))
            assert urls[miner_hotkey] == manifest_url

            reloaded_state = state_manager.get_state(reload=True)
            assert reloaded_state == state
        finally:
            shield.disable()
            assert not shield.run
            address_manager.clean_all()

    def test_ban_validator(self) -> None:
        """
        Test if shield is properly banning validators.
        """
        # Use long validate_interval_sec to avoid race conditions between manipulating data due to validation
        # and accessing state in tests.
        self.create_default_shield(validate_interval_sec=600)
        sleep(2)

        try:
            state: MinerShieldState = self.state_manager.get_state()
            assert state.known_validators == self.validators_manager.get_validators()
            assert state.banned_validators == {}

            # shield is working, ban single validator

            expected_validators: dict[Hotkey, PublicKey] = dict(self.validators_manager.get_validators())
            expected_validators.pop(self.VALIDATOR_1_HOTKEY)
            banned_validators: set[Hotkey] = {self.VALIDATOR_1_HOTKEY}
            self.shield.ban_validator(self.VALIDATOR_1_HOTKEY)
            sleep(2)

            state = self.state_manager.get_state()
            assert state.known_validators == expected_validators
            assert state.banned_validators.keys() == banned_validators
            assert self.manifest_manager.put_counter == 2
            assert self.blockchain_manager.put_counter == 1  # file is put under same address

            # check banning non-existing validator

            self.shield.ban_validator(self.OTHER_VALIDATOR_HOTKEY)
            banned_validators.add(self.OTHER_VALIDATOR_HOTKEY)
            sleep(2)

            state = self.state_manager.get_state()
            assert state.known_validators == expected_validators
            assert state.banned_validators.keys() == banned_validators
            assert self.manifest_manager.put_counter == 2  # known validators haven't changed
        finally:
            self.shield.disable()

    def test_reloading_validators(self) -> None:
        """
        Test if shield is properly handling changing validators set during runtime.
        """
        self.create_default_shield(validate_interval_sec=5)
        sleep(2)

        try:
            # add VALIDATOR_4 and remove VALIDATOR_2 - changes should be reflected in the state after validation runs
            self.validators_manager.validators.pop(self.VALIDATOR_2_HOTKEY)
            self.validators_manager.validators[self.VALIDATOR_4_HOTKEY] = self.VALIDATOR_4_PUBLICKEY
            expected_validators: dict[Hotkey, PublicKey] = dict(self.DEFAULT_VALIDATORS)
            expected_validators.pop(self.VALIDATOR_2_HOTKEY)
            expected_validators[self.VALIDATOR_4_HOTKEY] = self.VALIDATOR_4_PUBLICKEY
            state: MinerShieldState = self.state_manager.get_state()
            expected_address_id_counter: int = len(state.validators_addresses)
            assert self.address_manager.id_counter == expected_address_id_counter
            assert state.known_validators != expected_validators  # need to wait
            assert self.manifest_manager.put_counter == 1

            # wait for validation and check results
            sleep(self.shield.options.validate_interval_sec)
            state = self.state_manager.get_state()
            assert state.known_validators == expected_validators
            assert state.validators_addresses.keys() == state.known_validators.keys()
            assert list(state.validators_addresses.values()) == list(self.address_manager.known_addresses.values())
            expected_address_id_counter += 1  # +1 for new validator
            assert self.address_manager.id_counter == expected_address_id_counter
            assert self.manifest_manager.put_counter == 2

            # change public key of single validator - changes should be reflected in the state after validation runs
            self.validators_manager.validators[self.VALIDATOR_3_HOTKEY] = self.VALIDATOR_3_OTHER_PUBLICKEY
            expected_validators[self.VALIDATOR_3_HOTKEY] = self.VALIDATOR_3_OTHER_PUBLICKEY
            state = self.state_manager.get_state()
            assert state.known_validators != expected_validators  # need to wait

            # wait for validation and check results
            sleep(self.shield.options.validate_interval_sec)
            state = self.state_manager.get_state()
            assert state.known_validators == expected_validators
            assert self.address_manager.id_counter == expected_address_id_counter
            assert self.manifest_manager.put_counter == 3
        finally:
            self.shield.disable()

    def test_validate_addresses(self) -> None:
        """
        Test if shield is properly handling validating addresses during runtime.
        """
        self.create_default_shield(validate_interval_sec=5)
        sleep(2)

        try:
            expected_address_id_counter: int = len(self.DEFAULT_VALIDATORS)

            # set invalid address manually - new address should be created after validation runs
            self.address_manager.invalid_addresses = {self.VALIDATOR_1_HOTKEY}
            state: MinerShieldState = self.state_manager.get_state()
            assert self.address_manager.id_counter == expected_address_id_counter
            old_address = state.validators_addresses[self.VALIDATOR_1_HOTKEY]
            assert self.address_manager.known_addresses.get(old_address.address_id) == old_address

            # wait for validation and check results
            expected_address_id_counter += 1  # +1 for regenerating address
            sleep(self.shield.options.validate_interval_sec)
            state = self.state_manager.get_state()
            assert self.address_manager.id_counter == expected_address_id_counter
            assert self.manifest_manager.put_counter == 2
            new_address = state.validators_addresses[self.VALIDATOR_1_HOTKEY]
            assert old_address != new_address
            assert self.address_manager.known_addresses.get(new_address.address_id) == new_address
            assert self.address_manager.known_addresses.get(old_address.address_id, None) is None
        finally:
            self.shield.disable()

    def test_validate_manifest_file(self):
        """
        Test if shield is properly handling validating manifest file during runtime.
        """
        self.create_default_shield(validate_interval_sec=5)
        sleep(2)

        try:
            # change file contents manually - file should be overwritten after validation runs
            assert self.manifest_manager.put_counter == 1
            self.manifest_manager.stored_file = b'xxx'
            sleep(self.shield.options.validate_interval_sec)
            assert self.manifest_manager.put_counter == 2
        finally:
            self.shield.disable()



from __future__ import annotations

import asyncio
import time
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

import bittensor
from bittensor.core.extrinsics.serving import (
    publish_metadata,
    serve_extrinsic,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    import bittensor_wallet
    from bittensor.core.chain_data.neuron_info import NeuronInfo
    from scalecodec.base import ScaleType

    from bt_ddos_shield.event_processor import AbstractMinerShieldEventProcessor
    from bt_ddos_shield.utils import Hotkey, PublicKey


class BlockchainManagerException(Exception):
    pass


class AbstractBlockchainManager(ABC):
    """
    Abstract base class for manager handling publishing manifest address to blockchain.
    """

    def put_manifest_url(self, url: str):
        """
        Put manifest url to blockchain for wallet owner.
        """
        self.put_metadata(url.encode())

    async def get_manifest_urls(self, hotkeys: Iterable[Hotkey]) -> dict[Hotkey, str | None]:
        """
        Get manifest urls for given neurons identified by hotkeys.
        Returns dictionary with urls for given neurons, filled with None if url is not found.
        """
        try:
            serialized_urls: dict[Hotkey, bytes | None] = await self.get_metadata(hotkeys)
        except BlockchainManagerException:
            # Retry once
            time.sleep(3)
            serialized_urls = await self.get_metadata(hotkeys)
        deserialized_urls: dict[Hotkey, str | None] = {}
        for hotkey, serialized_url in serialized_urls.items():
            url: str | None = None
            if serialized_url is not None:
                try:
                    url = serialized_url.decode()
                except UnicodeDecodeError:
                    pass
            deserialized_urls[hotkey] = url
        return deserialized_urls

    async def get_own_manifest_url(self) -> str | None:
        """
        Get manifest url for wallet owner. Returns None if url is not found.
        """
        own_hotkey: Hotkey = self.get_hotkey()
        urls: dict[Hotkey, str | None] = await self.get_manifest_urls([own_hotkey])
        return urls.get(own_hotkey)

    @abstractmethod
    def put_metadata(self, data: bytes):
        """
        Put neuron metadata to blockchain for wallet owner.
        """
        pass

    @abstractmethod
    async def get_metadata(self, hotkeys: Iterable[Hotkey]) -> dict[Hotkey, bytes | None]:
        """
        Get metadata from blockchain for given neurons identified by hotkeys.
        Returns dictionary with metadata for given neurons, filled with None if metadata is not found.
        """
        pass

    @abstractmethod
    def get_hotkey(self) -> Hotkey:
        """Returns hotkey of the wallet owner."""
        pass

    @abstractmethod
    def get_own_public_key(self) -> PublicKey | None:
        """Returns public key for wallet owner."""
        pass

    @abstractmethod
    def upload_public_key(self, public_key: PublicKey):
        """Uploads public key to blockchain for wallet owner."""
        pass


class BittensorBlockchainManager(AbstractBlockchainManager):
    """
    Bittensor BlockchainManager implementation using commitments of knowledge as storage.
    """

    subtensor: bittensor.Subtensor
    netuid: int
    wallet: bittensor_wallet.Wallet
    event_processor: AbstractMinerShieldEventProcessor

    def __init__(
        self,
        subtensor: bittensor.Subtensor,
        netuid: int,
        wallet: bittensor_wallet.Wallet,
        event_processor: AbstractMinerShieldEventProcessor,
    ):
        self.subtensor = subtensor
        self.netuid = netuid
        self.wallet = wallet
        self.event_processor = event_processor

    async def get_metadata(self, hotkeys: Iterable[Hotkey]) -> dict[Hotkey, bytes | None]:
        try:
            async with bittensor.AsyncSubtensor(self.subtensor.chain_endpoint) as async_subtensor:
                tasks = [self.get_single_metadata(async_subtensor, hotkey) for hotkey in hotkeys]
                results: list[bytes | None] = await asyncio.gather(*tasks)
            return dict(zip(hotkeys, results, strict=True))
        except Exception as e:
            self.event_processor.event('Failed to get metadata for netuid={netuid}', exception=e, netuid=self.netuid)
            raise BlockchainManagerException(f'Failed to get metadata: {e}') from e

    async def get_single_metadata(self, async_subtensor: bittensor.AsyncSubtensor, hotkey: Hotkey) -> bytes | None:
        metadata: dict = await async_subtensor.substrate.query(
            module='Commitments',
            storage_function='CommitmentOf',
            params=[self.netuid, hotkey],
        )

        try:
            # This structure is hardcoded in bittensor publish_metadata function, but corresponding get_metadata
            # function does not use it, so we need to extract the value manually.

            # Commented parts of the code shows how this extraction should look if async version will work properly,
            # the same as sync version does. Now async version doesn't decode raw SCALE objects properly. I left this
            # code as it will be easier one day to restore this proper parsing.
            # fields: list[dict[str, str]] = metadata["info"]["fields"]
            fields = metadata['info']['fields']

            # As for now there is only one field in metadata. Field contains map from type of data to data itself.
            # field: dict[str, str] = fields[0]
            field = fields[0][0]

            # Find data of 'Raw' type.
            for data_type in field.keys():
                if data_type.startswith('Raw'):
                    break
            else:
                return None

            # Raw data is hex-encoded and prefixed with '0x'.
            # return bytes.fromhex(field[data_type][2:])
            return bytes(field[data_type][0])
        except TypeError:
            return None
        except LookupError:
            return None

    def put_metadata(self, data: bytes):
        try:
            publish_metadata(
                self.subtensor,
                self.wallet,
                self.netuid,
                data_type=f'Raw{len(data)}',
                data=data,
                wait_for_inclusion=True,
                wait_for_finalization=True,
            )
        except Exception as e:
            self.event_processor.event(
                'Failed to publish metadata for netuid={netuid}, wallet={wallet}',
                exception=e,
                netuid=self.netuid,
                wallet=str(self.wallet),
            )
            raise BlockchainManagerException(f'Failed to publish metadata: {e}') from e

    def get_hotkey(self) -> Hotkey:
        return self.wallet.hotkey.ss58_address

    def get_own_public_key(self) -> PublicKey | None:
        certificate: ScaleType = self.subtensor.query_subtensor(
            name='NeuronCertificates',
            params=[self.netuid, self.get_hotkey()],
        )
        cert_value: dict[str, Any] | None = certificate.serialize()
        if cert_value is None or 'public_key' not in cert_value:
            return None
        cert_type: str = format(cert_value['algorithm'], '02x')
        return cert_type + cert_value['public_key'][2:]  # public_key is prefixed with '0x'

    def upload_public_key(self, public_key: PublicKey):
        try:
            # As for now there is no method for uploading only certificate to Subtensor, so we need to use
            # serve_extrinsic function. Because of that we need to get current neuron info to not overwrite existing
            # data - if there is not existing data, we will use default dummy values.
            neuron: NeuronInfo | None = self.subtensor.get_neuron_for_pubkey_and_subnet(
                self.wallet.hotkey.ss58_address, netuid=self.netuid
            )
            new_ip: str = neuron.axon_info.ip if neuron is not None else '127.0.0.1'
            new_port: int = neuron.axon_info.port if neuron is not None else 1
            new_protocol: int = neuron.axon_info.protocol if neuron is not None else 0
            # We need to change any field, otherwise extrinsic will not be sent, so use placeholder1 (increased by 1
            # and modulo 256 as it is u8 field) to not modify any real data.
            new_placeholder1: int = (neuron.axon_info.placeholder1 + 1) % 256 if neuron is not None else 0
            # certificate param is of str type in library, but actually we need to pass bytes there
            certificate_data: bytes = bytes.fromhex(public_key)

            serve_extrinsic(
                self.subtensor,
                self.wallet,
                new_ip,
                new_port,
                new_protocol,
                self.netuid,
                certificate=certificate_data,  # type: ignore
                placeholder1=new_placeholder1,
                wait_for_inclusion=True,
                wait_for_finalization=True,
            )
        except Exception as e:
            self.event_processor.event(
                'Failed to upload public key for netuid={netuid}, wallet={wallet}',
                exception=e,
                netuid=self.netuid,
                wallet=str(self.wallet),
            )
            raise BlockchainManagerException(f'Failed to upload public key: {e}') from e



from __future__ import annotations

import copy
from types import MappingProxyType
from typing import TYPE_CHECKING

import pytest

from bt_ddos_shield.address_manager import (
    AbstractAddressManager,
    AwsAddressManager,
    AwsObjectTypes,
    AwsShieldedServerData,
    ShieldedServerLocation,
    ShieldedServerLocationType,
)
from bt_ddos_shield.event_processor import PrintingMinerShieldEventProcessor
from bt_ddos_shield.utils import AWSClientFactory, Hotkey, ShieldAddress
from tests.test_state_manager import MemoryMinerShieldStateManager

if TYPE_CHECKING:
    from bt_ddos_shield.state_manager import MinerShieldState
    from tests.conftest import ShieldTestSettings


def get_miner_location_from_credentials(
    location_type: ShieldedServerLocationType, test_settings: ShieldTestSettings
) -> ShieldedServerLocation:
    location_value: str
    if location_type == ShieldedServerLocationType.EC2_ID:
        location_value = test_settings.aws_miner_instance_id
    else:
        assert location_type == ShieldedServerLocationType.EC2_IP
        location_value = test_settings.aws_miner_instance_ip
    return ShieldedServerLocation(
        location_type=location_type, location_value=location_value, port=test_settings.miner_instance_port
    )


class MemoryAddressManager(AbstractAddressManager):
    id_counter: int
    known_addresses: dict[str, ShieldAddress]
    invalid_addresses: set[Hotkey]

    def __init__(self):
        self.id_counter = 0
        self.known_addresses = {}
        self.invalid_addresses = set()

    def clean_all(self):
        self.id_counter = 0
        self.known_addresses.clear()
        self.invalid_addresses.clear()

    def create_address(self, hotkey: Hotkey) -> ShieldAddress:
        address = ShieldAddress(
            address_id=str(self.id_counter),
            address=f'addr{self.id_counter}.com',
            port=80,
        )
        self.known_addresses[address.address_id] = address
        self.id_counter += 1
        return address

    def remove_address(self, address: ShieldAddress):
        self.known_addresses.pop(address.address_id, None)

    def validate_addresses(self, addresses: MappingProxyType[Hotkey, ShieldAddress]) -> set[Hotkey]:
        for hotkey in self.invalid_addresses:
            if hotkey in addresses:
                self.known_addresses.pop(addresses[hotkey].address_id, None)
        return self.invalid_addresses


class TestAddressManager:
    """
    Test suite for the address manager.
    """

    def create_aws_address_manager(
        self,
        test_settings: ShieldTestSettings,
        location_type: ShieldedServerLocationType,
        create_state_manager: bool = True,
    ) -> AwsAddressManager:
        miner_location: ShieldedServerLocation = get_miner_location_from_credentials(location_type, test_settings)
        if create_state_manager:
            self.state_manager: MemoryMinerShieldStateManager = MemoryMinerShieldStateManager()
        aws_client_factory: AWSClientFactory = AWSClientFactory(
            test_settings.aws_access_key_id, test_settings.aws_secret_access_key, test_settings.aws_region_name
        )
        return AwsAddressManager(
            aws_client_factory=aws_client_factory,
            server_location=miner_location,
            hosted_zone_id=test_settings.aws_route53_hosted_zone_id,
            event_processor=PrintingMinerShieldEventProcessor(),
            state_manager=self.state_manager,
        )

    @pytest.fixture
    def address_manager(self, shield_settings: ShieldTestSettings):
        manager = self.create_aws_address_manager(shield_settings, ShieldedServerLocationType.EC2_ID)
        yield manager
        manager.clean_all()

    def test_address_manager_creation(self, shield_settings: ShieldTestSettings):
        address_manager_for_ec2_id = self.create_aws_address_manager(shield_settings, ShieldedServerLocationType.EC2_ID)
        address_manager_for_ec2_ip = self.create_aws_address_manager(shield_settings, ShieldedServerLocationType.EC2_IP)
        assert address_manager_for_ec2_id.shielded_server_data == address_manager_for_ec2_ip.shielded_server_data, (
            'IP address should be resolved to instance ID'
        )

    def test_create_elb_for_ec2(self, shield_settings: ShieldTestSettings, address_manager: AwsAddressManager):
        """
        Test creating ELB by AwsAddressManager class when shielding EC2 instance.

        IMPORTANT: Test can run for many minutes due to AWS delays.
        """

        # This triggers creation of ELB
        address_manager.validate_addresses(MappingProxyType({}))

        state: MinerShieldState = self.state_manager.get_state()
        address_manager_state: MappingProxyType[str, str] = state.address_manager_state
        current_hosted_zone_id: str = address_manager_state[address_manager.HOSTED_ZONE_ID_STATE_KEY]
        assert current_hosted_zone_id == shield_settings.aws_route53_hosted_zone_id
        json_data: str = state.address_manager_state[address_manager.SHIELDED_SERVER_STATE_KEY]
        server_data = AwsShieldedServerData.from_json(json_data)
        assert server_data.aws_location and server_data.aws_location.server_id == shield_settings.aws_miner_instance_id
        assert server_data.server_location.port == shield_settings.miner_instance_port
        created_objects: MappingProxyType[str, frozenset[str]] = state.address_manager_created_objects
        assert len(created_objects[AwsObjectTypes.WAF.value]) == 1
        assert len(created_objects[AwsObjectTypes.ELB.value]) == 1
        assert AwsObjectTypes.VPC.value not in created_objects
        assert len(created_objects[AwsObjectTypes.SUBNET.value]) == 1
        assert len(created_objects[AwsObjectTypes.TARGET_GROUP.value]) == 1
        assert len(created_objects[AwsObjectTypes.SECURITY_GROUP.value]) == 1
        assert len(created_objects[AwsObjectTypes.DNS_ENTRY.value]) == 1

        reloaded_state: MinerShieldState = self.state_manager.get_state(reload=True)
        assert reloaded_state == state

        # Call clean_all and check if everything was removed
        address_manager.clean_all()
        state = self.state_manager.get_state()
        created_objects = state.address_manager_created_objects
        assert AwsObjectTypes.WAF.value not in created_objects
        assert AwsObjectTypes.ELB.value not in created_objects
        assert AwsObjectTypes.VPC.value not in created_objects
        assert AwsObjectTypes.SUBNET.value not in created_objects
        assert AwsObjectTypes.TARGET_GROUP.value not in created_objects
        assert AwsObjectTypes.SECURITY_GROUP.value not in created_objects
        assert AwsObjectTypes.DNS_ENTRY.value not in created_objects

    def test_handle_address(self, address_manager: AwsAddressManager):
        """
        Create address, validate it and remove created address.

        IMPORTANT: Test can run for many minutes due to AWS delays.
        """
        address1: ShieldAddress = address_manager.create_address('validator1')
        address2: ShieldAddress = address_manager.create_address('validator2')
        invalid_address: ShieldAddress = ShieldAddress(address_id='invalid', address='invalid.com', port=80)
        mapping: dict[Hotkey, ShieldAddress] = {'hotkey1': address1, 'hotkey2': address2, 'invalid': invalid_address}
        invalid_addresses: set[Hotkey] = address_manager.validate_addresses(MappingProxyType(mapping))
        assert invalid_addresses == {'invalid'}

        state: MinerShieldState = self.state_manager.get_state()
        created_objects: MappingProxyType[str, frozenset[str]] = state.address_manager_created_objects
        assert len(created_objects[AwsObjectTypes.ELB.value]) == 1, 'ELB should be created before adding address'
        assert len(created_objects[AwsObjectTypes.DNS_ENTRY.value]) == 1, 'only wildcard DNS entry should be created'

        address_manager.remove_address(address1)
        state = self.state_manager.get_state()
        created_objects = state.address_manager_created_objects
        assert len(created_objects[AwsObjectTypes.ELB.value]) == 1
        assert len(created_objects[AwsObjectTypes.DNS_ENTRY.value]) == 1
        invalid_addresses = address_manager.validate_addresses(MappingProxyType(mapping))
        assert invalid_addresses == {'hotkey1', 'invalid'}

    def test_miner_instance_change(self, shield_settings: ShieldTestSettings, address_manager: AwsAddressManager):
        """
        Test changing Miner instance when initializing shield.

        IMPORTANT: Test can run for many minutes due to AWS delays.
        """
        address: ShieldAddress = address_manager.create_address('validator1')
        hotkey: Hotkey = 'hotkey'
        mapping: dict[Hotkey, ShieldAddress] = {hotkey: address}
        invalid_addresses: set[Hotkey] = address_manager.validate_addresses(MappingProxyType(mapping))
        assert invalid_addresses == set()

        state: MinerShieldState = self.state_manager.get_state()
        created_objects: MappingProxyType[str, frozenset[str]] = state.address_manager_created_objects
        assert len(created_objects[AwsObjectTypes.ELB.value]) == 1, 'ELB should be created before adding address'
        elb_id: str = next(iter(created_objects[AwsObjectTypes.ELB.value]))
        hosted_zone_id: str = state.address_manager_state[address_manager.HOSTED_ZONE_ID_STATE_KEY]
        dns_entry_id: str = next(iter(created_objects[AwsObjectTypes.DNS_ENTRY.value]))

        # Create new manager with different port - ELB should be recreated
        new_test_settings: ShieldTestSettings = copy.deepcopy(shield_settings)
        new_test_settings.miner_instance_port += 1
        address_manager = self.create_aws_address_manager(
            new_test_settings, ShieldedServerLocationType.EC2_ID, create_state_manager=False
        )
        invalid_addresses = address_manager.validate_addresses(MappingProxyType(mapping))
        assert invalid_addresses == {hotkey}

        state = self.state_manager.get_state()
        created_objects = state.address_manager_created_objects
        assert len(created_objects[AwsObjectTypes.DNS_ENTRY.value]) == 1
        assert len(created_objects[AwsObjectTypes.ELB.value]) == 1
        new_elb_id: str = next(iter(created_objects[AwsObjectTypes.ELB.value]))
        assert new_elb_id != elb_id
        new_hosted_zone_id: str = state.address_manager_state[address_manager.HOSTED_ZONE_ID_STATE_KEY]
        assert hosted_zone_id == new_hosted_zone_id
        new_dns_entry_id: str = next(iter(created_objects[AwsObjectTypes.DNS_ENTRY.value]))
        assert dns_entry_id == new_dns_entry_id

    def test_hosted_zone_id_change(self, shield_settings: ShieldTestSettings, address_manager: AwsAddressManager):
        """
        Test changing hosted zone id when initializing shield.

        IMPORTANT: Test can run for many minutes due to AWS delays.
        """
        address: ShieldAddress = address_manager.create_address('validator1')
        hotkey: Hotkey = 'hotkey'
        mapping: dict[Hotkey, ShieldAddress] = {hotkey: address}
        invalid_addresses: set[Hotkey] = address_manager.validate_addresses(MappingProxyType(mapping))
        assert invalid_addresses == set()

        state: MinerShieldState = self.state_manager.get_state()
        created_objects: MappingProxyType[str, frozenset[str]] = state.address_manager_created_objects
        assert len(created_objects[AwsObjectTypes.DNS_ENTRY.value]) == 1
        assert len(created_objects[AwsObjectTypes.ELB.value]) == 1, 'ELB should be created before adding address'
        elb_id: str = next(iter(created_objects[AwsObjectTypes.ELB.value]))
        hosted_zone_id: str = state.address_manager_state[address_manager.HOSTED_ZONE_ID_STATE_KEY]
        dns_entry_id: str = next(iter(created_objects[AwsObjectTypes.DNS_ENTRY.value]))

        try:
            # Create new manager with different hosted zone - only addresses should be removed
            new_test_settings: ShieldTestSettings = copy.deepcopy(shield_settings)
            new_test_settings.aws_route53_hosted_zone_id = shield_settings.aws_route53_other_hosted_zone_id
            address_manager = self.create_aws_address_manager(
                new_test_settings, ShieldedServerLocationType.EC2_ID, create_state_manager=False
            )
            invalid_addresses = address_manager.validate_addresses(MappingProxyType(mapping))
            assert invalid_addresses == {hotkey}

            state = self.state_manager.get_state()
            created_objects = state.address_manager_created_objects
            assert len(created_objects[AwsObjectTypes.DNS_ENTRY.value]) == 1
            assert len(created_objects[AwsObjectTypes.ELB.value]) == 1
            new_elb_id: str = next(iter(created_objects[AwsObjectTypes.ELB.value]))
            assert new_elb_id == elb_id
            new_hosted_zone_id: str = state.address_manager_state[address_manager.HOSTED_ZONE_ID_STATE_KEY]
            assert hosted_zone_id != new_hosted_zone_id
            new_dns_entry_id: str = next(iter(created_objects[AwsObjectTypes.DNS_ENTRY.value]))
            assert dns_entry_id != new_dns_entry_id
        finally:
            address_manager.clean_all()  # Needed to remove created DNS entry in new hosted zone



from __future__ import annotations

from abc import ABC, abstractmethod
from types import MappingProxyType
from typing import TYPE_CHECKING, Any

from substrateinterface.base import QueryMapResult  # noqa

from bt_ddos_shield.encryption_manager import CertificateAlgorithmEnum

if TYPE_CHECKING:
    from collections.abc import Iterable

    import bittensor

    from bt_ddos_shield.utils import Hotkey, PublicKey


class AbstractValidatorsManager(ABC):
    """
    Abstract base class for manager of validators and their public keys used for encryption.
    """

    @abstractmethod
    def get_validators(self) -> MappingProxyType[Hotkey, PublicKey]:
        """
        Get cached dictionary of validators.

        Returns:
            dict[Hotkey, PublicKey]: Mapping HotKey of validator -> his public key.
        """
        pass

    @abstractmethod
    def reload_validators(self):
        """
        Reload validators dictionary. Blocks code execution until new validators set is fetched.
        """
        pass


class MemoryValidatorsManager(AbstractValidatorsManager):
    """
    Validators manager implementation which stores fixed validators in memory.
    """

    validators: dict[Hotkey, PublicKey]

    def __init__(self, validators: dict[Hotkey, PublicKey]):
        self.validators = dict(validators)

    def get_validators(self) -> MappingProxyType[Hotkey, PublicKey]:
        return MappingProxyType(self.validators)

    def reload_validators(self):
        pass


class BittensorValidatorsManager(AbstractValidatorsManager):
    """
    Validators Manager using Bittensor Neurons' Certificates.
    """

    # 1000 tao is needed to set weights and miners typically don't hold staked tao on their own hotkeys as it provides
    # no value. Therefore, we assume that any neuron with at least 1000 tao staked is a validator. It is simple
    # heuristic, which can be not valid for all subnets, but as for now is sufficient.
    MIN_VALIDATOR_STAKE = 1000

    subtensor: bittensor.Subtensor
    netuid: int
    validators: frozenset[Hotkey]
    certificates: dict[Hotkey, PublicKey]

    def __init__(
        self,
        subtensor: bittensor.Subtensor,
        netuid: int,
        validators: Iterable[Hotkey] | None = None,
    ):
        self.subtensor = subtensor
        self.netuid = netuid
        self.validators = frozenset(validators or [])
        self.certificates = {}

    def get_validators(self) -> MappingProxyType[Hotkey, PublicKey]:
        return MappingProxyType(self.certificates)

    def reload_validators(self) -> None:
        validators: frozenset[Hotkey]
        if self.validators:
            validators = self.validators
        else:
            neurons: list[bittensor.NeuronInfoLite] = self.subtensor.neurons_lite(self.netuid)
            validators = frozenset(neuron.hotkey for neuron in neurons if self.is_validator(neuron))

        self.certificates = self.fetch_certificates(validators)

    def fetch_certificates(self, validators: frozenset[Hotkey]) -> dict[Hotkey, PublicKey]:
        """
        Fetch Validators' Certificates (PublicKey) from Subtensor for given validators identified by hotkeys.
        """
        query_certificates: QueryMapResult = self.subtensor.query_map(
            module='SubtensorModule',
            name='NeuronCertificates',
            params=[self.netuid],
        )
        certificates: dict[Hotkey, dict[str, Any]] = {
            hotkey.serialize(): certificate.serialize() for hotkey, certificate in query_certificates
        }
        cert_type: str = format(CertificateAlgorithmEnum.ECDSA_SECP256K1_UNCOMPRESSED, '02x')
        return {
            hotkey: cert_type + certificate['public_key'][2:]  # Key is prefixed with '0x'
            for hotkey, certificate in certificates.items()
            if hotkey in validators
            and certificate['algorithm'] == CertificateAlgorithmEnum.ECDSA_SECP256K1_UNCOMPRESSED
        }

    def is_validator(self, neuron: bittensor.NeuronInfoLite) -> bool:
        """
        Determine whether provided Neuron is a Validator or not.
        """
        return neuron.stake >= self.MIN_VALIDATOR_STAKE



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime
from types import MappingProxyType
from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    DateTime,
    Engine,
    ForeignKey,
    Integer,
    PrimaryKeyConstraint,
    String,
    create_engine,
)
from sqlalchemy.orm import DeclarativeBase, sessionmaker

from bt_ddos_shield.utils import ShieldAddress

if TYPE_CHECKING:
    from sqlalchemy.engine import url

    from bt_ddos_shield.utils import Hotkey, PublicKey


class MinerShieldState:
    _known_validators: dict[Hotkey, PublicKey]
    _banned_validators: dict[Hotkey, datetime]
    _validators_addresses: dict[Hotkey, ShieldAddress]
    _address_manager_state: dict[str, str]
    _address_manager_created_objects: dict[str, frozenset[str]]

    def __init__(
        self,
        known_validators: dict[Hotkey, PublicKey] | None = None,
        banned_validators: dict[Hotkey, datetime] | None = None,
        validators_addresses: dict[Hotkey, ShieldAddress] | None = None,
        address_manager_state: dict[str, str] | None = None,
        address_manager_created_objects: dict[str, frozenset[str]] | None = None,
    ):
        super().__setattr__('_known_validators', known_validators or {})
        super().__setattr__('_banned_validators', banned_validators or {})
        super().__setattr__('_validators_addresses', validators_addresses or {})
        super().__setattr__('_address_manager_state', address_manager_state or {})
        super().__setattr__('_address_manager_created_objects', address_manager_created_objects or {})

    @property
    def known_validators(self) -> MappingProxyType[Hotkey, PublicKey]:
        """
        Get dictionary of known validators - maps validator HotKey -> validator public key.
        """
        return MappingProxyType(self._known_validators)

    @property
    def banned_validators(self) -> MappingProxyType[Hotkey, datetime]:
        """
        Get dictionary of banned validators - maps validator HotKey -> time of ban.
        """
        return MappingProxyType(self._banned_validators)

    @property
    def validators_addresses(self) -> MappingProxyType[Hotkey, ShieldAddress]:
        """
        Get dictionary of active addresses (validator HotKey -> Address created for him).
        """
        return MappingProxyType(self._validators_addresses)

    @property
    def address_manager_state(self) -> MappingProxyType[str, str]:
        """
        Get address manager state (key -> value).
        """
        return MappingProxyType(self._address_manager_state)

    @property
    def address_manager_created_objects(self) -> MappingProxyType[str, frozenset[str]]:
        """
        Get objects already created by address manager (object_type -> set of object_ids).
        """
        return MappingProxyType(self._address_manager_created_objects)

    def __setattr__(self, key, value):
        raise AttributeError('State is immutable')

    def __delattr__(self, item):
        raise AttributeError('State is immutable')

    def __eq__(self, other):
        if not isinstance(other, MinerShieldState):
            return False

        return (
            self._known_validators == other._known_validators
            and self._banned_validators == other._banned_validators
            and self._validators_addresses == other._validators_addresses
            and self._address_manager_state == other._address_manager_state
            and self._address_manager_created_objects == other._address_manager_created_objects
        )


class AbstractMinerShieldStateManager(ABC):
    """
    Abstract base class for manager handling state of MinerShield. Each change in state should be instantly
    saved to storage.
    """

    current_miner_shield_state: MinerShieldState = MinerShieldState()
    _initialized: bool = False

    def get_state(self, reload: bool = False) -> MinerShieldState:
        """
        Get current state of MinerShield. If state is not loaded, it is loaded first.
        """
        if reload or not self._initialized:
            self.current_miner_shield_state = self._load_state_from_storage()

        return self.current_miner_shield_state

    @abstractmethod
    def add_validator(self, validator_hotkey: Hotkey, validator_public_key: PublicKey, redirect_address: ShieldAddress):
        """
        Add validator together with his public key and address (created for him) redirecting to Miner server.
        """
        pass

    @abstractmethod
    def update_validator_public_key(self, validator_hotkey: Hotkey, validator_public_key: PublicKey):
        pass

    @abstractmethod
    def add_banned_validator(self, validator_hotkey: Hotkey):
        pass

    @abstractmethod
    def remove_validator(self, validator_hotkey: Hotkey):
        """
        Remove validator from the sets of known validators and active addresses.
        """
        pass

    @abstractmethod
    def update_address_manager_state(self, key: str, value: str | None):
        """
        Update address manager state (key -> value). If value is None, remove key from state.
        """
        pass

    @abstractmethod
    def add_address_manager_created_object(self, obj_type: str, obj_id: str):
        pass

    @abstractmethod
    def del_address_manager_created_object(self, obj_type: str, obj_id: str):
        pass

    @abstractmethod
    def _load_state_from_storage(self) -> MinerShieldState:
        pass

    def _update_state(
        self,
        known_validators: dict[Hotkey, PublicKey] | None = None,
        banned_validators: dict[Hotkey, datetime] | None = None,
        validators_addresses: dict[Hotkey, ShieldAddress] | None = None,
        address_manager_state: dict[str, str] | None = None,
        address_manager_created_objects: dict[str, frozenset[str]] | None = None,
    ):
        """
        Create new updated state with given new values and set it as current state. If value for field is None,
        it is copied from current state.
        """
        self.current_miner_shield_state = MinerShieldState(
            dict(self.current_miner_shield_state.known_validators) if known_validators is None else known_validators,
            dict(self.current_miner_shield_state.banned_validators) if banned_validators is None else banned_validators,
            dict(self.current_miner_shield_state.validators_addresses)
            if validators_addresses is None
            else validators_addresses,
            dict(self.current_miner_shield_state.address_manager_state)
            if address_manager_state is None
            else address_manager_state,
            dict(self.current_miner_shield_state.address_manager_created_objects)
            if address_manager_created_objects is None
            else address_manager_created_objects,
        )

    def _state_add_validator(
        self, validator_hotkey: Hotkey, validator_public_key: PublicKey, redirect_address: ShieldAddress
    ):
        """
        Add new validator to current state. Should be called only after updating state in storage.
        """
        known_validators: dict[Hotkey, PublicKey] = dict(self.current_miner_shield_state.known_validators)
        assert validator_hotkey not in known_validators, 'storage should not allow adding same validator'
        known_validators[validator_hotkey] = validator_public_key

        validators_addresses: dict[Hotkey, ShieldAddress] = dict(self.current_miner_shield_state.validators_addresses)
        assert validator_hotkey not in validators_addresses, 'storage should not allow adding same validator'
        validators_addresses[validator_hotkey] = redirect_address

        self._update_state(known_validators=known_validators, validators_addresses=validators_addresses)

    def _state_update_validator_public_key(self, validator_hotkey: Hotkey, validator_public_key: PublicKey):
        """
        Update validator in current state. Should be called only after updating state in storage.
        """
        known_validators: dict[Hotkey, PublicKey] = dict(self.current_miner_shield_state.known_validators)
        assert validator_hotkey in known_validators, 'updating storage should fail when validator does not exists'
        known_validators[validator_hotkey] = validator_public_key
        self._update_state(known_validators=known_validators)

    def _state_add_banned_validator(self, validator_hotkey: Hotkey, ban_time: datetime):
        """
        Add new banned validator to current state. Should be called only after updating state in storage.
        """
        banned_validators: dict[Hotkey, datetime] = dict(self.current_miner_shield_state.banned_validators)
        assert validator_hotkey not in banned_validators, 'time should be updated only when adding new ban'
        banned_validators[validator_hotkey] = ban_time
        self._update_state(banned_validators=banned_validators)

    def _state_remove_validator(self, validator_hotkey: Hotkey):
        """
        Remove validator from current state. Should be called only after updating state in storage.
        """
        known_validators: dict[Hotkey, PublicKey] = dict(self.current_miner_shield_state.known_validators)
        assert validator_hotkey in known_validators, 'storage should not allow removing non-existent validator'
        known_validators.pop(validator_hotkey)
        validators_addresses: dict[Hotkey, ShieldAddress] = dict(self.current_miner_shield_state.validators_addresses)
        assert validator_hotkey in validators_addresses, 'storage should not allow removing non-existent validator'
        validators_addresses.pop(validator_hotkey)
        self._update_state(known_validators=known_validators, validators_addresses=validators_addresses)

    def _state_update_address_manager_state(self, key: str, value: str | None):
        """
        Updates AddressManager state in current shield state. Should be called only after updating state in storage.
        """
        address_manager_state: dict[str, str] = dict(self.current_miner_shield_state.address_manager_state)
        if value is None:
            address_manager_state.pop(key, None)
        else:
            address_manager_state[key] = value
        self._update_state(address_manager_state=address_manager_state)

    def _state_add_address_manager_created_object(self, obj_type: str, obj_id: str):
        """
        Add object to objects created by AddressManager. Should be called only after updating state in storage.
        """
        address_manager_created_objects: dict[str, frozenset[str]] = dict(
            self.current_miner_shield_state.address_manager_created_objects
        )
        if obj_type not in address_manager_created_objects:
            address_manager_created_objects[obj_type] = frozenset([obj_id])
        else:
            address_manager_created_objects[obj_type] = address_manager_created_objects[obj_type] | frozenset([obj_id])
        self._update_state(address_manager_created_objects=address_manager_created_objects)

    def _state_del_address_manager_created_object(self, obj_type: str, obj_id: str):
        """
        Remove object from objects created by AddressManager. Should be called only after updating state in storage.
        """
        address_manager_created_objects: dict[str, frozenset[str]] = dict(
            self.current_miner_shield_state.address_manager_created_objects
        )
        if obj_type not in address_manager_created_objects:
            return
        address_manager_created_objects[obj_type] = frozenset(
            o for o in address_manager_created_objects[obj_type] if o != obj_id
        )
        if not address_manager_created_objects[obj_type]:
            address_manager_created_objects.pop(obj_type)
        self._update_state(address_manager_created_objects=address_manager_created_objects)


class MinerShieldStateDeclarativeBase(DeclarativeBase):
    pass


class SqlValidator(MinerShieldStateDeclarativeBase):
    __tablename__ = 'validators'
    hotkey = Column(String, primary_key=True)
    public_key = Column(String, nullable=False)
    address_id = Column(String, ForeignKey('addresses.address_id', ondelete='CASCADE'), nullable=False)


class SqlAddress(MinerShieldStateDeclarativeBase):
    __tablename__ = 'addresses'
    address_id = Column(String, primary_key=True)
    address = Column(String, nullable=False)
    port = Column(Integer, nullable=False)


class SqlBannedValidator(MinerShieldStateDeclarativeBase):
    __tablename__ = 'banned_validators'
    hotkey = Column(String, primary_key=True)
    ban_time = Column(DateTime, nullable=False)


class SqlAddressManagerState(MinerShieldStateDeclarativeBase):
    __tablename__ = 'address_manager_state'
    key = Column(String, primary_key=True)
    value = Column(String, nullable=False)


class SqlAddressManagerCreatedObjects(MinerShieldStateDeclarativeBase):
    __tablename__ = 'address_manager_created_objects'
    object_type = Column(String, nullable=False)
    object_id = Column(String, nullable=False)
    __table_args__ = (PrimaryKeyConstraint('object_type', 'object_id', name='pk_object_type_object_id'),)


class SQLAlchemyMinerShieldStateManager(AbstractMinerShieldStateManager):
    """
    StateManager implementation using SQLAlchemy.
    """

    engine: Engine
    session_maker: sessionmaker

    def __init__(self, db_url: str | url.URL):
        """
        Args:
            db_url: URL of database to connect to. Should be in format accepted by SQLAlchemy - see create_engine doc.
        """
        super().__init__()
        self.engine = create_engine(db_url)
        MinerShieldStateDeclarativeBase.metadata.create_all(self.engine)
        self.session_maker = sessionmaker(bind=self.engine)

    def clear_tables(self):
        MinerShieldStateDeclarativeBase.metadata.drop_all(self.engine)
        MinerShieldStateDeclarativeBase.metadata.create_all(self.engine)

    def add_validator(self, validator_hotkey: Hotkey, validator_public_key: PublicKey, redirect_address: ShieldAddress):
        with self.session_maker() as session:
            session.add(
                SqlValidator(
                    hotkey=validator_hotkey, public_key=validator_public_key, address_id=redirect_address.address_id
                )
            )
            session.add(
                SqlAddress(
                    address_id=redirect_address.address_id,
                    address=redirect_address.address,
                    port=redirect_address.port,
                )
            )
            session.commit()

        self._state_add_validator(validator_hotkey, validator_public_key, redirect_address)

    def update_validator_public_key(self, validator_hotkey: Hotkey, validator_public_key: PublicKey):
        with self.session_maker() as session:
            validator = session.query(SqlValidator).filter_by(hotkey=validator_hotkey).one()
            validator.public_key = validator_public_key
            session.commit()

        self._state_update_validator_public_key(validator_hotkey, validator_public_key)

    def add_banned_validator(self, validator_hotkey: Hotkey):
        if validator_hotkey in self.current_miner_shield_state.banned_validators:
            # do not update ban time
            return

        ban_time: datetime = datetime.now()

        with self.session_maker() as session:
            session.add(SqlBannedValidator(hotkey=validator_hotkey, ban_time=ban_time))
            session.commit()

        self._state_add_banned_validator(validator_hotkey, ban_time)

    def remove_validator(self, validator_hotkey: Hotkey):
        with self.session_maker() as session:
            validator = session.query(SqlValidator).filter_by(hotkey=validator_hotkey).one()
            session.delete(validator)
            session.commit()

        self._state_remove_validator(validator_hotkey)

    def update_address_manager_state(self, key: str, value: str | None):
        with self.session_maker() as session:
            if value is None:
                # Remove the key from the database if the value is None
                session.query(SqlAddressManagerState).filter_by(key=key).delete()
            else:
                # Insert or update the key-value pair in the database
                state = session.query(SqlAddressManagerState).filter_by(key=key).one_or_none()
                if state is None:
                    session.add(SqlAddressManagerState(key=key, value=value))
                else:
                    state.value = value
            session.commit()

        self._state_update_address_manager_state(key, value)

    def add_address_manager_created_object(self, obj_type: str, obj_id: str):
        with self.session_maker() as session:
            session.add(SqlAddressManagerCreatedObjects(object_type=obj_type, object_id=obj_id))
            session.commit()

        self._state_add_address_manager_created_object(obj_type, obj_id)

    def del_address_manager_created_object(self, obj_type: str, obj_id: str):
        with self.session_maker() as session:
            session.query(SqlAddressManagerCreatedObjects).filter_by(object_type=obj_type, object_id=obj_id).delete()
            session.commit()

        self._state_del_address_manager_created_object(obj_type, obj_id)

    def _load_state_from_storage(self) -> MinerShieldState:
        with self.session_maker() as session:
            # noinspection PyTypeChecker
            known_validators: dict[Hotkey, PublicKey] = {
                v.hotkey: v.public_key for v in session.query(SqlValidator).all()
            }
            # noinspection PyTypeChecker
            banned_validators: dict[Hotkey, datetime] = {
                b.hotkey: b.ban_time for b in session.query(SqlBannedValidator).all()
            }
            # noinspection PyTypeChecker
            validators_addresses: dict[Hotkey, ShieldAddress] = {
                v.hotkey: self._load_address(session, v.address_id) for v in session.query(SqlValidator).all()
            }

            # noinspection PyTypeChecker
            address_manager_state: dict[str, str] = {
                s.key: s.value for s in session.query(SqlAddressManagerState).all()
            }

            tmp_address_manager_created_objects: defaultdict[str, set[str]] = defaultdict(set)
            for obj in session.query(SqlAddressManagerCreatedObjects).all():
                # noinspection PyTypeChecker
                tmp_address_manager_created_objects[obj.object_type].add(obj.object_id)

        address_manager_created_objects: dict[str, frozenset[str]] = {}
        for obj_type in tmp_address_manager_created_objects:
            address_manager_created_objects[obj_type] = frozenset(tmp_address_manager_created_objects[obj_type])

        return MinerShieldState(
            known_validators,
            banned_validators,
            validators_addresses,
            address_manager_state,
            address_manager_created_objects,
        )

    @classmethod
    def _load_address(cls, session, address_id: str) -> ShieldAddress:
        db_address = session.query(SqlAddress).filter_by(address_id=address_id).one()
        return ShieldAddress(
            address_id=db_address.address_id,
            address=db_address.address,
            port=db_address.port,
        )



from __future__ import annotations

import traceback
from abc import ABC, abstractmethod
from datetime import datetime
from sys import stdout
from typing import Any


class MinerShieldEvent:
    """
    Class describing event, which happened in the shield.
    """

    date: datetime  # Date of the event.
    template: str  # Template of the event.
    exception: Exception | None  # Exception which caused the event.
    metadata: dict[str, Any]  # Additional metadata.
    description: str  # Description of the event (template filled with metadata).

    def __init__(self, template: str, exception: Exception | None = None, **metadata):
        """
        Args:
            template: Template of the event (used to generate description).
            exception: Exception which caused the event.
            metadata: Additional metadata (used also to generate description).
        """
        self.date = datetime.now()
        self.template = template
        self.exception = exception
        self.metadata = metadata
        self.description = template.format(**metadata)


class AbstractMinerShieldEventProcessor(ABC):
    """
    Abstract base class for processor handling events generated by shield.
    """

    def event(self, template: str, exception: Exception | None = None, **kwargs):
        """
        Add event to event processor.

        Args:
            template: Description template to be filled using kwargs.
            exception: Exception to be attached to event.
            kwargs: Event params. Used also to fill template.
        """
        return self._add_event(MinerShieldEvent(template, exception, **kwargs))

    @abstractmethod
    def _add_event(self, event: MinerShieldEvent):
        pass


class PrintingMinerShieldEventProcessor(AbstractMinerShieldEventProcessor):
    """
    Event processor which logs events to console.
    """

    def _add_event(self, event: MinerShieldEvent):
        print(f'{event.date}: MinerShieldEvent: {event.description}\nMetadata: {event.metadata}')
        if event.exception is not None:
            print('Exception happened:')
            traceback.print_exception(event.exception, file=stdout)



from __future__ import annotations

from typing import TYPE_CHECKING

from bt_ddos_shield.miner_shield import MinerShield, MinerShieldFactory
from bt_ddos_shield.shield_metagraph import ShieldMetagraph
from bt_ddos_shield.state_manager import SQLAlchemyMinerShieldStateManager
from bt_ddos_shield.validators_manager import BittensorValidatorsManager

if TYPE_CHECKING:
    from bittensor.core.chain_data import AxonInfo

    from tests.conftest import ShieldTestSettings


class TestValidator:
    """
    Test suite for the Validator class.
    """

    def test_full_flow(self, shield_settings: ShieldTestSettings):
        """
        Test if validator is working using real managers and real shield.

        IMPORTANT: Test can run for many minutes due to AWS delays.
        """
        miner_hotkey: str = shield_settings.wallet.instance.hotkey.ss58_address

        # We need to add any validator to set, otherwise manifest addresses will be created for all validators in
        # network including tested validator, what we don't want
        validators = {'unknown_hotkey'}
        shield: MinerShield = MinerShieldFactory.create_miner_shield(shield_settings, validators)

        assert isinstance(shield.state_manager, SQLAlchemyMinerShieldStateManager)
        state_manager: SQLAlchemyMinerShieldStateManager = shield.state_manager
        state_manager.clear_tables()

        assert isinstance(shield.validators_manager, BittensorValidatorsManager)
        validators_manager: BittensorValidatorsManager = shield.validators_manager

        shield.enable()
        assert shield.run
        shield.task_queue.join()  # Wait for full shield initialization - should create empty manifest

        try:
            metagraph: ShieldMetagraph = ShieldMetagraph(
                wallet=shield_settings.validator_wallet.instance,
                certificate_path=shield_settings.validator_cert_path,
                subtensor=shield_settings.subtensor.create_client(),
                netuid=shield_settings.netuid,
            )
            miner_axon: AxonInfo = next(axon for axon in metagraph.axons if axon.hotkey == miner_hotkey)

            shield.disable()
            validators_manager.validators = frozenset()
            shield.enable()
            shield.task_queue.join()  # Wait for full shield initialization - should add validator to manifest

            metagraph.sync()
            shielded_miner_axon: AxonInfo = next(axon for axon in metagraph.axons if axon.hotkey == miner_hotkey)
            assert shielded_miner_axon.ip != miner_axon.ip
            assert shielded_miner_axon.port == shield_settings.miner_instance_port
        finally:
            shield.disable()
            assert not shield.run
            shield.address_manager.clean_all()



from __future__ import annotations

import enum
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, NamedTuple, TypeVar

import ecies
from coincurve.keys import PrivateKey as CoincurvePrivateKey

if TYPE_CHECKING:
    from bt_ddos_shield.utils import PrivateKey, PublicKey


class EncryptionManagerException(Exception):
    pass


class EncryptionError(EncryptionManagerException):
    pass


class DecryptionError(EncryptionManagerException):
    pass


class EncryptionCertificate(NamedTuple):
    private_key: PrivateKey
    public_key: PublicKey


CertType = TypeVar('CertType')


class AbstractEncryptionManager[CertType](ABC):
    """
    Abstract base class for manager handling manifest file encryption.
    """

    @abstractmethod
    def encrypt(self, public_key: PublicKey, data: bytes) -> bytes:
        """
        Encrypts given data using the provided public key. Throws EncryptionError if encryption fails.
        """
        pass

    @abstractmethod
    def decrypt(self, private_key: PrivateKey, data: bytes) -> bytes:
        """
        Decrypts given data using the provided private key. Throws DecryptionError if decryption fails.
        """
        pass

    @classmethod
    @abstractmethod
    def generate_certificate(cls) -> CertType:
        """
        Generates certificate object, which will be used for encryption of manifest data.
        """
        pass

    @classmethod
    @abstractmethod
    def serialize_certificate(cls, certificate: CertType) -> EncryptionCertificate:
        """
        Serialize certificate public and private key.
        """
        pass

    @classmethod
    @abstractmethod
    def save_certificate(cls, certificate: CertType, path: str):
        """
        Save certificate to disk.
        """
        pass

    @classmethod
    @abstractmethod
    def load_certificate(cls, path: str) -> CertType:
        """
        Load certificate from disk.
        """
        pass


class CertificateAlgorithmEnum(enum.IntEnum):
    """Values are taken from coincurve.keys.PublicKey.__init__ method."""

    ECDSA_SECP256K1_UNCOMPRESSED = 4
    """ ECDSA using secp256k1 curve (uncompressed version) """


class ECIESEncryptionManager(AbstractEncryptionManager[CoincurvePrivateKey]):
    """
    Encryption manager implementation using ECIES algorithm. Public and private keys are Coincurve (secp256k1) keys
    in hex format.
    """

    def encrypt(self, public_key: PublicKey, data: bytes) -> bytes:
        try:
            return ecies.encrypt(public_key, data)
        except Exception as e:
            raise EncryptionError(f'Encryption failed: {e}') from e

    def decrypt(self, private_key: PrivateKey, data: bytes) -> bytes:
        try:
            return ecies.decrypt(private_key, data)
        except Exception as e:
            raise DecryptionError(f'Decryption failed: {e}') from e

    @classmethod
    def generate_certificate(cls) -> CoincurvePrivateKey:
        return ecies.utils.generate_key()

    @classmethod
    def serialize_certificate(cls, certificate: CoincurvePrivateKey) -> EncryptionCertificate:
        private_key: str = certificate.to_hex()
        public_key: bytes = certificate.public_key.format(compressed=False)
        assert public_key[0] == CertificateAlgorithmEnum.ECDSA_SECP256K1_UNCOMPRESSED
        return EncryptionCertificate(private_key, public_key.hex())

    @classmethod
    def save_certificate(cls, certificate: CoincurvePrivateKey, path: str):
        with open(path, 'wb') as f:
            f.write(certificate.to_pem())

    @classmethod
    def load_certificate(cls, path: str) -> CoincurvePrivateKey:
        with open(path, 'rb') as f:
            return CoincurvePrivateKey.from_pem(f.read())
